from pathlib import Path

from unimport import utils
from unimport.color import difference as color_difference

__all__ = ("diff",)


def diff(
    path: Path,
    source: str,
    refactor_result: str,
) -> bool:
    diff_ = utils.diff(
        source=source, refactor_result=refactor_result, fromfile=path
    )
    exists_diff = bool(diff_)
    if exists_diff:
        print(color_difference(diff_))

    return exists_diff



import argparse
from pathlib import Path

from unimport import constants as C
from unimport.config import DefaultConfig

__all__ = (
    "add_sources_option",
    "add_check_option",
    "add_config_option",
    "add_include_option",
    "add_exclude_option",
    "add_gitignore_option",
    "add_ignore_init_option",
    "add_include_star_import_option",
    "add_diff_option",
    "add_remove_option",
    "add_permission_option",
    "add_requirements_option",
    "add_version_option",
)

default_config = DefaultConfig()


def add_sources_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "sources",
        default=default_config.sources,
        nargs="*",
        help="Files and folders to find the unused imports.",
        action="store",
        type=Path,
    )


def add_check_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--check",
        action="store_true",
        help="Prints which file the unused imports are in.",
        default=default_config.check,
    )


def add_config_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-c",
        "--config",
        default=".",
        help="Read configuration from PATH.",
        metavar="PATH",
        action="store",
        type=Path,
    )


def add_include_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--include",
        help="File include pattern.",
        metavar="include",
        action="store",
        default=default_config.include,
        type=str,
    )


def add_exclude_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--exclude",
        help="File exclude pattern.",
        metavar="exclude",
        action="store",
        default=default_config.exclude,
        type=str,
    )


def add_gitignore_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--gitignore",
        action="store_true",
        help="Exclude .gitignore patterns. if present.",
        default=default_config.gitignore,
    )


def add_ignore_init_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--ignore-init",
        action="store_true",
        help="Ignore the __init__.py file.",
        default=default_config.ignore_init,
    )


def add_include_star_import_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--include-star-import",
        action="store_true",
        help="Include star imports during scanning and refactor.",
        default=default_config.include_star_import,
    )


def add_diff_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-d",
        "--diff",
        action="store_true",
        help="Prints a diff of all the changes unimport would make to a file.",
        default=default_config.diff,
    )


def add_remove_option(exclusive_group: argparse._MutuallyExclusiveGroup):
    exclusive_group.add_argument(
        "-r",
        "--remove",
        action="store_true",
        help="Remove unused imports automatically.",
        default=default_config.remove,
    )


def add_permission_option(exclusive_group: argparse._MutuallyExclusiveGroup):
    exclusive_group.add_argument(
        "-p",
        "--permission",
        action="store_true",
        help="Refactor permission after see diff.",
        default=default_config.permission,
    )


def add_requirements_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--requirements",
        action="store_true",
        help="Include requirements.txt file, You can use it with all other arguments",
        default=default_config.requirements,
    )


def add_version_option(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"Unimport {C.VERSION}",
        help="Prints version of unimport",
    )



from pathlib import Path

from unimport.color import CYAN, GREEN, paint

__all__ = ("remove", "requirements_remove")


def remove(
    path: Path, encoding, newline, refactor_result: str, use_color_setting
) -> None:
    with open(path, mode="w", encoding=encoding, newline=newline) as py_file:
        py_file.write(refactor_result)

    print(f"Refactoring '{paint(str(path), GREEN, use_color_setting)}'")


def requirements_remove(
    path: Path, refactor_result: str, use_color_setting: bool
) -> None:
    path.write_text(refactor_result)
    print(f"Refactoring '{paint(path.as_posix(), CYAN, use_color_setting)}'")



import ast
import operator
import sys
from dataclasses import dataclass, field
from typing import ClassVar, Iterator, List, Set, Union

if sys.version_info >= (3, 8):
    from typing import Literal  # unimport: skip
else:
    from typing_extensions import Literal


__all__ = ("Import", "ImportFrom", "Name", "Scope")


@dataclass
class Import:
    imports: ClassVar[List[Union["Import", "ImportFrom"]]] = []

    lineno: int
    column: int
    name: str
    package: str

    node: ast.AST = field(init=False, repr=False, compare=False)

    def __len__(self) -> int:
        return operator.length_hint(self.name.split("."))

    def is_match_sub_packages(self, name_name) -> bool:
        return self.name.split(".")[0] == name_name

    @property
    def scope(self):
        return Scope.get_scope_by_current_node(self)

    def is_used(self) -> bool:
        for name in self.scope.names:
            if name is None:
                continue

            if name.match_import:
                if name.match_import == self:
                    return True
            elif name.match(self):
                return True

        return False

    def match_nearest_duplicate_import(self, name: "Name") -> bool:
        nearest_import = None

        scope = name.scope
        while scope:
            imports = [
                _import
                for _import in scope.imports
                if name.match_2(_import) and name.lineno > _import.lineno
            ]
            scope = scope.parent

            if imports:
                nearest_import = max(
                    filter(
                        lambda _import: _import.lineno
                        == max(
                            imports, key=lambda _import: _import.lineno
                        ).lineno,
                        imports,
                    ),
                    key=lambda _import: _import.column,
                )

            if nearest_import == self:
                return True

        return False

    @property
    def is_duplicate(self) -> bool:
        return [_import.name for _import in self.imports].count(self.name) > 1

    @classmethod
    def get_unused_imports(
        cls, include_star_import: bool = False
    ) -> Iterator[Union["Import", "ImportFrom"]]:
        for imp in reversed(Import.imports):
            if (
                include_star_import
                and isinstance(imp, ImportFrom)
                and imp.star
            ):
                yield imp
            elif not imp.is_used():
                yield imp

    @classmethod
    def register(
        cls, lineno: int, column: int, name: str, package: str, node: ast.AST
    ) -> None:
        _import = cls(lineno, column, name, package)
        _import.node = node
        cls.imports.append(_import)

        Scope.register(_import)

    @classmethod
    def clear(cls):
        cls.imports.clear()


@dataclass
class ImportFrom(Import):
    star: bool
    suggestions: List[str]

    def is_match_sub_packages(self, name_name):
        return False

    @classmethod
    def register(  # type: ignore
        cls,
        lineno: int,
        column: int,
        name: str,
        package: str,
        node: ast.AST,
        star: bool,
        suggestions: List[str],
    ) -> None:
        _import = cls(lineno, column, name, package, star, suggestions)
        _import.node = node
        cls.imports.append(_import)

        Scope.register(_import)


@dataclass
class Name:
    names: ClassVar[List["Name"]] = []

    lineno: int
    name: str
    is_all: bool = False

    node: ast.AST = field(init=False, repr=False, compare=False)
    match_import: Union[Import, Literal[False]] = field(
        init=False, repr=False, compare=False, default=False
    )

    @property
    def is_attribute(self):
        return "." in self.name

    def match_2(self, imp: Union[Import, ImportFrom]) -> bool:
        if self.is_all:
            is_match = self.name == imp.name
        elif self.is_attribute:
            is_match = imp.lineno < self.lineno and (
                ".".join(self.name.split(".")[: len(imp)]) == imp.name
                or imp.is_match_sub_packages(self.name)
            )
        else:
            is_match = (imp.lineno < self.lineno) and (
                self.name == imp.name or imp.is_match_sub_packages(self.name)
            )

        return is_match

    def match(self, imp: Union[Import, ImportFrom]) -> bool:
        is_match = self.match_2(imp)

        if is_match and imp.is_duplicate:
            is_match = imp.match_nearest_duplicate_import(self)

        if is_match:
            self.match_import = imp

        return is_match

    @property
    def scope(self):
        return Scope.get_scope_by_current_node(self)

    @classmethod
    def register(
        cls, lineno: int, name: str, node: ast.AST, is_all: bool = False
    ) -> None:
        _name = cls(lineno, name, is_all)
        _name.node = node
        cls.names.append(_name)

        Scope.register(_name, is_global=is_all)

    @classmethod
    def clear(cls) -> None:
        cls.names.clear()


@dataclass
class Scope:
    scopes: ClassVar[List["Scope"]] = []
    current_scope: ClassVar[List["Scope"]] = []

    node: ast.AST

    current_nodes: List[Union[Import, ImportFrom, Name]] = field(
        default_factory=list, init=False, repr=False, compare=False
    )
    parent: "Scope" = field(default=None, repr=False)
    child_scopes: Set["Scope"] = field(
        default_factory=set, init=False, repr=False, compare=False
    )

    def __hash__(self) -> int:
        return hash(self.node)

    @classmethod
    def get_curent_scope(cls) -> "Scope":
        return cls.current_scope[-1]

    @classmethod
    def get_global_scope(cls) -> "Scope":
        global_scope = cls.scopes[0]
        assert global_scope.parent is None
        return global_scope

    @classmethod
    def add_global_scope(cls, tree: ast.AST) -> None:
        parent = None
        scope = Scope(tree, parent)
        cls.current_scope.append(scope)
        cls.scopes.append(scope)  # global scope added to cls.scopes

    @classmethod
    def add_current_scope(cls, node: ast.AST) -> None:
        parent = cls.get_curent_scope()
        scope = Scope(node, parent)
        cls.current_scope.append(scope)

    @classmethod
    def remove_current_scope(cls):
        cls.current_scope.pop()

    @classmethod
    def register(
        cls, current_node: Union[Import, ImportFrom, Name], *, is_global=False
    ) -> None:
        scope = cls.get_previous_scope(
            cls.get_global_scope() if is_global else cls.get_curent_scope()
        )

        # current nodes add to scope
        scope.current_nodes.append(current_node)

        # child scopes add to scope
        if scope.parent is None:
            return

        parent = cls.get_previous_scope(scope.parent)
        child_scope = scope

        while parent:
            parent.child_scopes.add(child_scope)

            child_scope = parent
            if parent.parent is None:
                break
            parent = cls.get_previous_scope(parent.parent)

    @classmethod
    def get_scope_by_current_node(
        cls, current_node: Union[Import, ImportFrom, Name]
    ) -> "Scope":
        for scope in cls.scopes:
            if current_node in scope.current_nodes:
                return scope
        return None

    @property
    def names(self) -> Iterator[Name]:
        yield from filter(  # type: ignore
            lambda node: isinstance(node, Name), self.current_nodes
        )

        for child_scope in self.child_scopes:
            yield from child_scope.names

    @property
    def imports(self) -> Iterator[Import]:
        yield from filter(  # type: ignore
            lambda node: isinstance(node, Import), self.current_nodes
        )

    @classmethod
    def get_previous_scope(cls, scope: "Scope") -> "Scope":
        for _scope in cls.scopes:
            if _scope == scope:
                return _scope

        cls.scopes.append(scope)
        return scope

    @classmethod
    def clear(cls):
        cls.scopes.clear()



from pathlib import Path
from typing import List, Union

from unimport.color import CYAN, GREEN, RED, YELLOW, paint
from unimport.statement import Import, ImportFrom

__all__ = ("check", "requirements_check")


def check(
    path: Path,
    unused_imports: List[Union[Import, ImportFrom]],
    use_color_setting: bool,
) -> None:
    for imp in unused_imports:
        if isinstance(imp, ImportFrom) and imp.star and imp.suggestions:
            context = (
                paint(f"from {imp.name} import *", RED, use_color_setting)
                + " -> "
                + paint(
                    f"from {imp.name} import {', '.join(imp.suggestions)}",
                    GREEN,
                    use_color_setting,
                )
            )
        else:
            context = paint(imp.name, YELLOW, use_color_setting)
        print(
            context
            + " at "
            + paint(path.as_posix(), GREEN, use_color_setting)
            + ":"
            + paint(str(imp.lineno), GREEN, use_color_setting)
        )


def requirements_check(
    path: Path, index: int, requirement: str, use_color_setting: bool
) -> None:
    print(
        f"{paint(requirement, CYAN, use_color_setting)} at "
        f"{paint(path.as_posix(), CYAN, use_color_setting)}:{paint(str(index + 1), CYAN, use_color_setting)}"
    )



from pathlib import Path

from unimport import utils
from unimport.color import CYAN, YELLOW, paint

__all__ = ("permission", "requirements_permission")


def permission(
    path: Path,
    encoding: str,
    newline: str,
    refactor_result: str,
    use_color_setting: bool,
) -> None:
    action = input(
        f"Apply suggested changes to '{paint(str(path), YELLOW, use_color_setting)}' [Y/n/q] ? >"
    ).lower()
    if action == "q":
        raise SystemExit(1)
    elif utils.actiontobool(action):
        from unimport.commands import remove

        remove(path, encoding, newline, refactor_result, use_color_setting)


def requirements_permission(
    path: Path, refactor_result: str, use_color_setting: bool
):
    action = input(
        f"Apply suggested changes to '{paint(path.as_posix(), CYAN, use_color_setting)}' [Y/n/q] ? >"
    ).lower()
    if action == "q":
        return 1
    if utils.actiontobool(action):
        from unimport.commands import requirements_remove

        requirements_remove(path, refactor_result, use_color_setting)



import argparse
import sys
from typing import Optional, Sequence, Set

from unimport import color, commands
from unimport import constants as C
from unimport import emoji, utils
from unimport.analyzer import Analyzer
from unimport.commands import options
from unimport.config import Config
from unimport.refactor import refactor_string
from unimport.statement import Import

__all__ = ("main",)


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = argparse.ArgumentParser(
        prog="unimport",
        description=C.DESCRIPTION,
        epilog=f"Get rid of all unused imports {emoji.PARTYING_FACE}",
    )
    exclusive_group = parser.add_mutually_exclusive_group(required=False)

    options.add_sources_option(parser)
    options.add_check_option(parser)
    options.add_config_option(parser)
    color.add_color_option(parser)
    options.add_include_option(parser)
    options.add_exclude_option(parser)
    options.add_gitignore_option(parser)
    options.add_ignore_init_option(parser)
    options.add_include_star_import_option(parser)
    options.add_diff_option(parser)
    options.add_remove_option(exclusive_group)
    options.add_permission_option(exclusive_group)
    options.add_requirements_option(parser)
    options.add_version_option(parser)

    argv = argv if argv is not None else sys.argv[1:]
    args = parser.parse_args(argv)
    config = Config.get_config(args)

    unused_modules = set()
    used_packages: Set[str] = set()

    for path in config.get_paths():
        source, encoding, newline = utils.read(path)

        with Analyzer(
            source=source,
            path=path,
            include_star_import=config.include_star_import,
        ):
            unused_imports = list(
                Import.get_unused_imports(config.include_star_import)
            )
            unused_modules.update({imp.name for imp in unused_imports})
            used_packages.update(
                utils.get_used_packages(Import.imports, unused_imports)
            )
            if config.check:
                commands.check(path, unused_imports, args.color)
            if any((config.diff, config.remove)):
                refactor_result = refactor_string(
                    source=source,
                    unused_imports=unused_imports,
                )
                if config.diff:
                    exists_diff = commands.diff(path, source, refactor_result)
                if config.permission and exists_diff:
                    commands.permission(
                        path, encoding, newline, refactor_result, args.color
                    )
                if config.remove and source != refactor_result:
                    commands.remove(
                        path, encoding, newline, refactor_result, args.color
                    )

    if not unused_modules and config.check:
        print(
            color.paint(
                f"{emoji.STAR} Congratulations there is no unused import in your project. {emoji.STAR}",
                color.GREEN,
                args.color,
            )
        )
    if config.requirements:
        for path in config.get_requirements():
            source = path.read_text()
            copy_source = source.splitlines().copy()

            for index, requirement in enumerate(source.splitlines()):
                module_name = utils.package_name_from_metadata(
                    requirement.split("==")[0]
                )
                if module_name is None:
                    print(
                        color.paint(
                            requirement + " not found", color.RED, args.color
                        )
                    )
                    continue

                if module_name not in used_packages:
                    copy_source.remove(requirement)

                    if config.check:
                        commands.requirements_check(
                            path, index, requirement, args.color
                        )

            refactor_result = "\n".join(copy_source)
            if config.diff:
                exists_diff = commands.diff(path, source, refactor_result)
                if config.permission and exists_diff:
                    commands.requirements_permission(
                        path, refactor_result, args.color
                    )
                if config.remove and source != refactor_result:
                    commands.requirements_remove(
                        path, refactor_result, args.color
                    )

    if unused_modules:
        return 1
    else:
        return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""This module performs static analysis using AST on the python code that's
given as a string and reports its findings."""

import ast
import functools
import re
from pathlib import Path
from typing import FrozenSet, List, Set, cast

from unimport import color
from unimport import constants as C
from unimport import utils
from unimport.relate import first_occurrence, get_parents, relate
from unimport.statement import Import, ImportFrom, Name, Scope

__all__ = ("Analyzer",)


def _generic_visit(func: C.FunctionT) -> C.FunctionT:
    """decorator to make visitor work _generic_visit."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)
        obj = args[0]
        node = args[1]
        obj.generic_visit(node)

    return cast(C.FunctionT, wrapper)


class _DefinedNameAnalyzer(ast.NodeVisitor):
    __slots__ = ("defined_names",)

    def __init__(self):
        self.defined_names: Set[str] = set()

    @_generic_visit
    def visit_FunctionDef(self, node: C.ASTFunctionT) -> None:
        self.defined_names.add(node.name)

    visit_AsyncFunctionDef = visit_FunctionDef

    @_generic_visit
    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.defined_names.add(node.name)

    @_generic_visit
    def visit_Name(self, node: ast.Name) -> None:
        if isinstance(node.ctx, ast.Store):
            self.defined_names.add(node.id)


class _ImportAnalyzer(ast.NodeVisitor):
    __slots__ = (
        "source",
        "include_star_import",
        "any_import_error",
        "defined_names",
    )

    ignore_modules_imports = ("__future__",)
    skip_import_comments_regex = "#.*(unimport: {0,1}skip|noqa)"

    def __init__(
        self,
        *,
        source: str,
        include_star_import: bool = False,
    ):
        self.source = source
        self.include_star_import = include_star_import

        self.any_import_error = False
        self.defined_names: Set[str] = set()

    def traverse(self, tree) -> None:
        defined_name_scanner = _DefinedNameAnalyzer()
        defined_name_scanner.visit(tree)
        self.defined_names = defined_name_scanner.defined_names

        self.visit(tree)

    def visit_ClassDef(self, node) -> None:
        Scope.add_current_scope(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    def visit_FunctionDef(self, node: C.ASTFunctionT) -> None:
        Scope.add_current_scope(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_AsyncFunctionDef = visit_FunctionDef

    @_generic_visit
    def visit_Import(self, node: ast.Import) -> None:
        if self.skip_import(node):
            return None

        for column, alias in enumerate(node.names):
            Import.register(
                lineno=node.lineno,
                column=column + 1,
                name=(alias.asname or alias.name),
                package=alias.name,
                node=node,
            )

    @_generic_visit
    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if self.skip_import(node):
            return None

        is_star = node.names[0].name == "*"
        for column, alias in enumerate(node.names):
            if not node.level:
                package = node.module
            else:
                package = "." * node.level + str(node.module) or ""
            alias_name = alias.asname or alias.name
            if (package in self.ignore_modules_imports) or (
                is_star and not self.include_star_import
            ):
                return
            ImportFrom.register(
                lineno=node.lineno,
                column=column + 1,
                name=package if is_star else alias_name,
                star=is_star,
                suggestions=self.get_suggestions(package) if is_star else [],
                package=package,
                node=node,
            )

    def visit_Try(self, node: ast.Try) -> None:
        self.any_import_error = True

        self.generic_visit(node)

        self.any_import_error = False

    def skip_import(self, node: C.ASTImport) -> bool:
        if C.PY38_PLUS:
            source_segment = "\n".join(
                self.source.splitlines()[node.lineno - 1 : node.end_lineno]
            )
        else:
            source_segment = self.source.splitlines()[node.lineno - 1]
        return (
            bool(
                re.search(
                    self.skip_import_comments_regex,
                    source_segment,
                    re.IGNORECASE,
                )
            )
            or self.any_import_error
        )

    def get_suggestions(self, package: str) -> List[str]:
        names = set(map(lambda name: name.name.split(".")[0], Name.names))
        from_names = _ImportableAnalyzer.get_names(package)
        return sorted(from_names & (names - self.defined_names))


class _NameAnalyzer(ast.NodeVisitor):
    def visit_ClassDef(self, node) -> None:
        Scope.add_current_scope(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    def visit_FunctionDef(self, node: C.ASTFunctionT) -> None:
        Scope.add_current_scope(node)

        self._type_comment(node)
        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_str_helper(self, value: str, node: ast.AST) -> None:
        parent = first_occurrence(node, *C.ASTFunctionTuple)
        is_annassign_or_arg = any(
            isinstance(parent, (ast.AnnAssign, ast.arg))
            for parent in get_parents(node)
        )
        if is_annassign_or_arg or (
            parent is not None and parent.returns is node
        ):
            self.join_visit(value, node)

    def visit_Str(self, node: ast.Str) -> None:
        self.visit_str_helper(node.s, node)

    def visit_Constant(self, node: ast.Constant) -> None:
        if isinstance(node.value, str):
            self.visit_str_helper(node.value, node)

    @_generic_visit
    def visit_Name(self, node: ast.Name) -> None:
        if not isinstance(node.parent, ast.Attribute):  # type: ignore
            Name.register(lineno=node.lineno, name=node.id, node=node)

    @_generic_visit
    def visit_Attribute(self, node: ast.Attribute) -> None:
        if not isinstance(node.value, ast.Call):
            names = []
            for sub_node in ast.walk(node):
                if isinstance(sub_node, ast.Attribute):
                    names.append(sub_node.attr)
                elif isinstance(sub_node, ast.Name):
                    names.append(sub_node.id)
            names.reverse()
            Name.register(lineno=node.lineno, name=".".join(names), node=node)

    @_generic_visit
    def visit_Assign(self, node: ast.Assign) -> None:
        self._type_comment(node)

    @_generic_visit
    def visit_arg(self, node: ast.arg) -> None:
        self._type_comment(node)

    @_generic_visit
    def visit_Subscript(self, node: ast.Subscript) -> None:
        # type_variable
        # type_var = List["object"] etc.

        def visit_constant_str(node: C.ASTNameType) -> None:
            """Separates the value by node type (str or constant) and gives it
            to the visit function."""

            if isinstance(node, ast.Constant) and isinstance(node.value, str):
                self.join_visit(node.value, node)
            elif isinstance(node, ast.Str):
                self.join_visit(node.s, node)

        if (
            isinstance(node.value, ast.Attribute)
            and isinstance(node.value.value, ast.Name)
            and node.value.value.id == "typing"
        ) or (
            isinstance(node.value, ast.Name)
            and node.value.id in C.SUBSCRIPT_TYPE_VARIABLE
        ):

            if C.PY39_PLUS:
                _slice = node.slice
            else:
                _slice = node.slice.value  # type: ignore

            if isinstance(_slice, ast.Tuple):  # type: ignore
                for elt in _slice.elts:  # type: ignore
                    if isinstance(elt, (ast.Constant, ast.Str)):
                        visit_constant_str(elt)
            else:
                if isinstance(_slice, (ast.Constant, ast.Str)):  # type: ignore
                    visit_constant_str(_slice)  # type: ignore

    @_generic_visit
    def visit_Call(self, node: ast.Call) -> None:
        # type_variable
        # cast("type", return_value)
        if (
            (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "typing"
                and node.func.attr == "cast"
            )
            or isinstance(node.func, ast.Name)
            and node.func.id == "cast"
        ):
            if isinstance(node.args[0], ast.Constant) and isinstance(
                node.args[0].value, str
            ):
                self.join_visit(node.args[0].value, node.args[0])
            elif isinstance(node.args[0], ast.Str):
                self.join_visit(node.args[0].s, node.args[0])

    def _type_comment(self, node: ast.AST) -> None:
        if isinstance(node, C.ASTFunctionTuple):
            mode = "func_type"
        else:
            mode = "eval"
        type_comment = getattr(node, "type_comment", None)
        if type_comment is not None:
            self.join_visit(type_comment, node, mode=mode)

    def join_visit(
        self, value: str, node: ast.AST, *, mode: str = "eval"
    ) -> None:
        """A function that parses the value, copies locations from the node and
        includes them in self.visit."""
        try:
            if C.PY38_PLUS:
                tree = ast.parse(value, mode=mode, type_comments=True)
            else:
                tree = ast.parse(value, mode=mode)
        except SyntaxError:
            return None
        else:
            relate(tree, parent=node.parent)  # type: ignore
            for new_node in ast.walk(tree):
                ast.copy_location(new_node, node)
            self.visit(tree)


class _ImportableAnalyzer(ast.NodeVisitor):
    __slots__ = (
        "importable_nodes",
        "suggestions_nodes",
    )

    def __init__(self) -> None:
        self.importable_nodes: List[
            C.ASTNameType
        ] = []  # nodes on the __all__ list
        self.suggestions_nodes: List[C.ASTImportableT] = []  # nodes on the CFN

    def visit_CFN(self, node: C.CFNT) -> None:
        Scope.add_current_scope(node)

        if not first_occurrence(node, C.DefTuple):
            self.suggestions_nodes.append(node)

        self.generic_visit(node)

        Scope.remove_current_scope()

    visit_ClassDef = visit_CFN
    visit_FunctionDef = visit_CFN
    visit_AsyncFunctionDef = visit_CFN

    @_generic_visit
    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            self.suggestions_nodes.append(alias)

    @_generic_visit
    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if not node.names[0].name == "*":
            for alias in node.names:
                self.suggestions_nodes.append(alias)

    @_generic_visit
    def visit_Assign(self, node: ast.Assign) -> None:
        if getattr(node.targets[0], "id", None) == "__all__" and isinstance(
            node.value, (ast.List, ast.Tuple, ast.Set)
        ):
            for item in node.value.elts:
                if isinstance(item, (ast.Constant, ast.Str)):
                    self.importable_nodes.append(item)

        for target in node.targets:  # we only get assigned names
            if isinstance(target, (ast.Name, ast.Attribute)):
                self.suggestions_nodes.append(target)

    @_generic_visit
    def visit_Expr(self, node: ast.Expr) -> None:
        if (
            isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Attribute)
            and isinstance(node.value.func.value, ast.Name)
            and node.value.func.value.id == "__all__"
        ):
            if node.value.func.attr == "append":
                for arg in node.value.args:
                    if isinstance(arg, (ast.Constant, ast.Str)):
                        self.importable_nodes.append(arg)

            elif node.value.func.attr == "extend":
                for arg in node.value.args:
                    if isinstance(arg, ast.List):
                        for item in arg.elts:
                            if isinstance(item, (ast.Constant, ast.Str)):
                                self.importable_nodes.append(item)

    @classmethod
    def get_names(cls, package: str) -> FrozenSet[str]:
        if utils.is_std(package):
            return utils.get_dir(package)

        source = utils.get_source(package)
        if source:
            try:
                tree = ast.parse(source)
            except SyntaxError:
                return frozenset()
            else:
                visitor = cls()
                relate(tree)
                visitor.visit(tree)
                return visitor.get_all() or visitor.get_suggestion()
        return frozenset()

    def get_all(self) -> FrozenSet[str]:
        names = set()
        for node in self.importable_nodes:
            if isinstance(node, ast.Constant):
                names.add(node.value)
            elif isinstance(node, ast.Str):
                names.add(node.s)
        return frozenset(names)

    def get_suggestion(self) -> FrozenSet[str]:
        names = set()
        for node in self.suggestions_nodes:  # type: ignore
            if isinstance(node, ast.Name):
                names.add(node.id)
            elif isinstance(node, ast.alias):
                names.add(node.asname or node.name)
            elif isinstance(node, C.DefTuple):
                names.add(node.name)
        return frozenset(names)

    def clear(self):
        self.importable_nodes.clear()
        self.suggestions_nodes.clear()


class Analyzer(ast.NodeVisitor):
    __slots__ = ("source", "path", "include_star_import")

    skip_file_regex = "#.*(unimport: {0,1}skip_file)"

    def __init__(
        self,
        *,
        source: str,
        path: Path = Path("<unknown file>"),
        include_star_import: bool = False,
    ):
        self.source = source
        self.path = path
        self.include_star_import = include_star_import

    def __enter__(self):
        self.traverse()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.clear()

    def traverse(self) -> None:
        if self.skip_file():
            return None

        try:
            if C.PY38_PLUS:
                tree = ast.parse(self.source, type_comments=True)
            else:
                tree = ast.parse(self.source)
        except SyntaxError as e:
            print(
                color.paint(str(e), color.RED)
                + " at "
                + color.paint(self.path.as_posix(), color.GREEN)
            )
            return None

        """
        Set parent
        """
        relate(tree)

        Scope.add_global_scope(tree)

        """
        Name analyzer
        """
        _NameAnalyzer().visit(tree)
        """
        Receive items on the __all__ list
        """
        importable_visitor = _ImportableAnalyzer()
        importable_visitor.visit(tree)
        for node in importable_visitor.importable_nodes:
            if isinstance(node, ast.Constant):
                Name.register(
                    lineno=node.lineno,
                    name=str(node.value),
                    node=node,
                    is_all=True,
                )
            elif isinstance(node, ast.Str):
                Name.register(
                    lineno=node.lineno, name=node.s, node=node, is_all=True
                )
        importable_visitor.clear()
        """
        Import analyzer
        """
        _ImportAnalyzer(
            source=self.source,
            include_star_import=self.include_star_import,
        ).traverse(tree)

        Scope.remove_current_scope()

    def skip_file(self) -> bool:
        return bool(
            re.search(self.skip_file_regex, self.source, re.IGNORECASE)
        )

    @staticmethod
    def clear():
        Name.clear()
        Import.clear()
        ImportFrom.clear()
        Scope.clear()



import sys
from pathlib import Path

from setuptools import setup

assert sys.version_info >= (3, 6), "unimport requires Python 3.6+"

CURRENT_DIR = Path(__file__).parent


def get_long_description():
    readme_md = CURRENT_DIR / "README.md"
    with open(readme_md, encoding="utf8") as ld_file:
        return ld_file.read()


DESCRIPTION = (
    "A linter, formatter for finding and removing unused import statements."
)
VERSION = "0.9.5"


setup(
    name="unimport",
    version=VERSION,
    description=DESCRIPTION,
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    keywords=["unused", "import"],
    author="Hakan Çelik",
    author_email="hakancelik96@outlook.com",
    url="https://github.com/hakancelik96/unimport",
    project_urls={
        "Documentation": "https://unimport.hakancelik.dev/",
        "Issues": "https://github.com/hakancelik96/unimport/issues",
        "Changelog": "https://unimport.hakancelik.dev/changelog/",
    },
    license="MIT",
    license_file="LICENSE",
    python_requires=">=3.6",
    packages=["unimport", "unimport.commands"],
    install_requires=[
        "libcst>=0.3.7; python_version >= '3.9'",
        "libcst>=0.3.0; python_version <= '3.8'",
        "pathspec>=0.5.0",
        "toml>=0.9.0",
        "importlib_metadata>=0.12; python_version < '3.8'",
        "dataclasses>=0.5; python_version < '3.7'",
        "typing-extensions>=3.7.4; python_version < '3.8'",
    ],
    extras_require={
        "docs": [
            "mkdocs==1.2.1",
            "mkdocs-material==7.1.9",
            "mkdocs-markdownextradata-plugin==0.2.4",
            "mkdocs-minify-plugin==0.4.0",
            "mkdocs-git-revision-date-localized-plugin==0.9.2",
        ],
        "dev": [
            "pytest==6.2.4",
            "pytest-cov==2.12.1",
            "semantic-version==2.8.5",
        ],
    },
    zip_safe=True,
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    entry_points={"console_scripts": ["unimport = unimport.__main__:main"]},
)



from unimport.commands.check import check, requirements_check
from unimport.commands.diff import diff
from unimport.commands.permission import permission, requirements_permission
from unimport.commands.remove import remove, requirements_remove

__all__ = (
    "check",
    "diff",
    "permission",
    "remove",
    "requirements_check",
    "requirements_permission",
    "requirements_remove",
)
